#include <glib.h>
#include <glib/gstdio.h>

#include <hildon/hildon.h>
#include <hildon/hildon-file-chooser-dialog.h>
#include <location/location-gpsd-control.h>
#include <location/location-gps-device.h>
#include <location/location-misc.h>
#include <location/location-distance-utils.h>

typedef struct {
    HildonProgram *program;
    HildonWindow *window;
    
    GtkWidget * main_vbox, *btn_hbox;
    GtkButton *start_stop_button, *save_button;
    GtkWidget *wp_hbox, *wp_label, *wp_name_entry;
    GtkWidget *loc_hbox, *loc_gps_data, *loc_track_data;
    GtkTable *loc_gps_data_table, *loc_track_data_table;
    GtkWidget *lat_val_label, *lon_val_label, *alt_val_label, *sat_val_label;
    GtkWidget *speed_val_label, *track_val_label, *climb_val_label;
    //GtkWidget *wp_label;
    GtkButton *wp_set_btn;
    GString *wp_marker_str, *waypoint_block_str;
    gboolean tracking_is_on, has_fix;
    FILE *outf_p;
    guint points_recorded_in_current_segment;
    LocationGPSDeviceStatus last_device_status;
    gchar *working_dir;
    gchar *intermediate_gpx_data_filename;
} AppData;

static gchar * interface_file_chooser (AppData * appdata, GtkFileChooserAction action)
{
    GtkWidget *dialog;
    gchar *filename = NULL;
    gchar tmpname[PATH_MAX];
    time_t t;
    struct tm *tmp;
    
    t = time(NULL);
    tmp = localtime(&t);
    strftime(tmpname, sizeof(tmpname), "gps-tracker-%Y%m%d_%H%M%S.gpx", tmp);
    dialog = hildon_file_chooser_dialog_new (GTK_WINDOW (appdata->window), action);
    gtk_widget_show_all (GTK_WIDGET (dialog));
    gtk_file_chooser_set_current_name(GTK_FILE_CHOOSER (dialog), tmpname);
    gtk_file_chooser_set_current_folder(GTK_FILE_CHOOSER(dialog), appdata->working_dir);

    if (gtk_dialog_run (GTK_DIALOG (dialog)) == GTK_RESPONSE_OK) {
        filename = gtk_file_chooser_get_filename (GTK_FILE_CHOOSER (dialog));
    }

    gtk_widget_destroy(dialog);
    return filename;
}

static void write_gpx_header(AppData *app_data)
{
  FILE *fp = app_data->outf_p;
  g_return_if_fail(fp);
  g_fprintf(fp,
      "<?xml version=\"1.0\" encoding=\"UTF-8\"?>\n"
      "<gpx version=\"1.0\"\n"
      "creator=\"GPS Træcker for Maemo\"\n"
      "xmlns:xsi=\"http://www.w3.org/2001/XMLSchema-instance\"\n"
      "xmlns=\"http://www.topografix.com/GPX/1/0\"\n"
      "xsi:schemaLocation=\"http://www.topografix.com/GPX/1/0 http://www.topografix.com/GPX/1/0/gpx.xsd\">\n"
      "<trk>\n"
      "<name>gps-tracker</name>\n"
      "<desc>This is a track recorded by 'GPS Træcker' on Maemo</desc>\n"
      "<trkseg>\n"
      );
}

static void write_gpx_footer(AppData *app_data)
{
  FILE *fp = app_data->outf_p;
  g_return_if_fail(fp);
  g_fprintf(fp, "</trkseg>\n</trk>\n");
  g_fprintf(fp, app_data->waypoint_block_str->str);
  g_fprintf(fp, "</gpx>\n");
}

static void cb_wp_set_btn (GtkWidget * w, AppData * data)
{
  HildonEntry *entry = (HildonEntry*)data->wp_name_entry;
  const gchar *marker_name = hildon_entry_get_text(entry);
  if(marker_name[0]) {
    g_string_assign(data->wp_marker_str, marker_name);
    //hildon_banner_show_information(GTK_WIDGET(data->window), NULL, data->wp_marker_str->str);
    hildon_entry_set_text(entry, "");
  }
  else {
    g_string_truncate(data->wp_marker_str, 0);
    hildon_banner_show_information(GTK_WIDGET(data->window), NULL, "WARNING: no marker name given");
  }
}

static void cb_start_stop (GtkWidget * w, AppData * data)
{
  data->tracking_is_on = !data->tracking_is_on;
  if(data->tracking_is_on) { /* START pressed */
    data->outf_p = fopen(data->intermediate_gpx_data_filename, "w");
    data->points_recorded_in_current_segment = 0;
    data->last_device_status = LOCATION_GPS_DEVICE_STATUS_NO_FIX;
    write_gpx_header(data);
    //hildon_banner_show_information(GTK_WIDGET(data->window), NULL, "Tracking started");
    gtk_button_set_label (data->start_stop_button, "Stop");
    gtk_widget_set_sensitive(GTK_WIDGET(data->save_button), FALSE);
    gtk_widget_set_sensitive(GTK_WIDGET(data->wp_set_btn), TRUE);
  }
  else { /* STOP pressed */
    data->points_recorded_in_current_segment = 0;
    data->last_device_status = LOCATION_GPS_DEVICE_STATUS_NO_FIX;
    if(data->outf_p) {
      write_gpx_footer(data);
      fclose(data->outf_p);
      data->outf_p = NULL;
    }
    gtk_button_set_label (data->start_stop_button, "Start");
    //hildon_banner_show_information(GTK_WIDGET(data->window), NULL, "Tracking stopped");
    gtk_widget_set_sensitive(GTK_WIDGET(data->save_button), TRUE);
    gtk_widget_set_sensitive(GTK_WIDGET(data->wp_set_btn), FALSE);
  }
  g_string_truncate(data->wp_marker_str, 0);
  g_string_truncate(data->waypoint_block_str, 0);
}

static void cb_file_save (GtkWidget * w, AppData * data)
{
    gchar *filename = NULL;
    filename = interface_file_chooser (data, GTK_FILE_CHOOSER_ACTION_SAVE);

    if (filename) {
      if(g_file_test(data->intermediate_gpx_data_filename, G_FILE_TEST_IS_REGULAR | G_FILE_TEST_EXISTS)) {
        g_rename(data->intermediate_gpx_data_filename, filename);
        //g_print ("File saved as %s\n", filename);
        hildon_banner_show_information(GTK_WIDGET(data->window), NULL, filename);
        gtk_widget_set_sensitive(GTK_WIDGET(data->save_button), FALSE);
      }
      else {
        hildon_banner_show_information(GTK_WIDGET(data->window), NULL, "Temp file not found");
      }
    }
    g_free(filename);
}

#define SPAN_PREFIX "<span font_family=\"monospace\" size=\"large\" weight=\"bold\">"
#define SPAN_POST   "</span>"
#define LABEL_FMT(s) SPAN_PREFIX s SPAN_POST
#define CHANGE_LABEL(w,fmt,val) change_label_text(w, SPAN_PREFIX fmt SPAN_POST, val)
#define CHANGE_LABEL2(w,fmt,val1, val2) change_label_text(w, SPAN_PREFIX fmt SPAN_POST, val1, val2)

static void change_label_text(GtkWidget *l, gchar *format, ...)
{
  va_list args;
  gchar buf[512];
  va_start(args, format);
  g_vsnprintf(buf, sizeof(buf), format, args);
  va_end(args);
  gtk_label_set_markup(GTK_LABEL(l), buf);
}

static void on_gps_device_changed (LocationGPSDevice *device, gpointer data)
{
  AppData *app_data = data;
  FILE *fp = app_data->outf_p;
  gchar sbuf1[G_ASCII_DTOSTR_BUF_SIZE], sbuf2[G_ASCII_DTOSTR_BUF_SIZE];
	if (!device)
		return;

  /* if there is no fix, or we are recording (fp != NULL), but the fix is so
   * bad that it is a 2D fix only, with no time set then do not
   * record any longer */
  if ((device->status == LOCATION_GPS_DEVICE_STATUS_NO_FIX) ||
      (device->fix && fp &&
       (! device->fix->fields & LOCATION_GPS_DEVICE_TIME_SET) &&
       (device->fix->mode == LOCATION_GPS_DEVICE_MODE_2D))) {
    app_data->has_fix = FALSE;
    //hildon_banner_show_information(GTK_WIDGET(app_data->window), NULL, "WARNING: Lost fix");
    app_data->last_device_status = LOCATION_GPS_DEVICE_STATUS_NO_FIX;
  }
  else if (device->fix) {
    //g_print("mode=%d stat=%d\n", device->fix->mode, device->status);
    if (device->fix->fields & LOCATION_GPS_DEVICE_LATLONG_SET) {
      app_data->has_fix = TRUE; /* fix found again */
      //g_print ("lat = %f, long = %f\n", device->fix->latitude, device->fix->longitude);
      CHANGE_LABEL(app_data->lat_val_label, "% 9.6f", device->fix->latitude);
      CHANGE_LABEL(app_data->lon_val_label, "% 9.6f", device->fix->longitude);
      if(fp) {
        if(app_data->last_device_status != device->status) {
          if(app_data->points_recorded_in_current_segment)
            g_fprintf(fp, "\n</trkseg>\n<trkseg>\n");
          app_data->last_device_status = device->status;
        }
        g_fprintf(fp, "\n<trkpt lat=\"%s\" lon=\"%s\">\n",
            g_ascii_formatd(sbuf1, sizeof(sbuf1), "%f", device->fix->latitude),
            g_ascii_formatd(sbuf2, sizeof(sbuf2), "%f", device->fix->longitude));
        if(app_data->wp_marker_str->len) {
          g_fprintf(fp, "<name>%s</name>\n", app_data->wp_marker_str->str);
          hildon_banner_show_information(GTK_WIDGET(app_data->window), NULL, app_data->wp_marker_str->str);
          g_string_append_printf(app_data->waypoint_block_str, "<wpt lat=\"%s\" lon=\"%s\"><name>%s</name></wpt>\n",
              sbuf1, sbuf2, app_data->wp_marker_str->str);
          g_string_truncate(app_data->wp_marker_str, 0);
        }
      }

      app_data->points_recorded_in_current_segment++;

      if (device->fix->fields & LOCATION_GPS_DEVICE_TIME_SET) {
        //g_print ("time = %f\n", device->fix->time);
        if(fp) {
          gchar st[64];
          struct tm *tmp;
          time_t t = device->fix->time;
          tmp = localtime(&t);
          strftime(st, sizeof(st), "%FT%T%Z", tmp);
          /* FIXME Is this really correct */
          g_fprintf(fp, "<time>%s</time>\n", st);
        }
      }
      if(fp) {
        if(device->fix->mode == LOCATION_GPS_DEVICE_MODE_2D)
          g_fprintf(fp, "<fix>2d</fix>\n");
        else if(device->fix->mode == LOCATION_GPS_DEVICE_MODE_3D)
          g_fprintf(fp, "<fix>3d</fix>\n");
        g_fprintf(fp, "<sat>%d</sat>\n", device->satellites_in_use);
      }

      if (device->fix->fields & LOCATION_GPS_DEVICE_ALTITUDE_SET) {
        //g_print ("alt = %f\n", device->fix->altitude);
        CHANGE_LABEL(app_data->alt_val_label, "% 5.0fm", device->fix->altitude);
        if(fp)
          g_fprintf(fp, "<ele>%s</ele>\n", g_ascii_formatd(sbuf1, sizeof(sbuf1), "%.1f", device->fix->altitude));

      }
      else
        CHANGE_LABEL(app_data->alt_val_label, " ?", NULL);

      if (device->fix->fields & LOCATION_GPS_DEVICE_SPEED_SET) {
        //g_print ("speed = %f\n", device->fix->speed);
        CHANGE_LABEL(app_data->speed_val_label, "% 5.1fkm/h", device->fix->speed);
        if(fp)
          g_fprintf(fp, "<speed>%s</speed>\n", g_ascii_formatd(sbuf1, sizeof(sbuf1), "%.1f", device->fix->speed));
      }
      else
        CHANGE_LABEL(app_data->speed_val_label, " ?", NULL);

      if (device->fix->fields & LOCATION_GPS_DEVICE_TRACK_SET) {
        //g_print ("track = %f\n", device->fix->track);
        CHANGE_LABEL(app_data->track_val_label, "%4.0f°", device->fix->track);
        if(fp)
          g_fprintf(fp, "<course>%s</course>\n", g_ascii_formatd(sbuf1, sizeof(sbuf1), "%.1f", device->fix->track));
      }
      else
        CHANGE_LABEL(app_data->track_val_label, " ?", NULL);

      if (device->fix->fields & LOCATION_GPS_DEVICE_CLIMB_SET) {
        //g_print ("climb = %f\n", device->fix->climb);
        CHANGE_LABEL(app_data->climb_val_label, "%+6.2fm/s", device->fix->climb);
      }
      else
        CHANGE_LABEL(app_data->climb_val_label, " ?", NULL);

      //g_print ("Accuracy values:\n");
      //g_print ("\tept = %e, eph = %e, epv = %e, epd = %e, "
      //    "eps = %e, epc = %e\n",
      //    device->fix->ept,
      //    device->fix->eph,
      //    device->fix->epv,
      //    device->fix->epd,
      //    device->fix->eps,
      //    device->fix->epc);
      if(fp) {
        g_fprintf(fp, "<hdop>%s</hdop>\n<vdop>%s</vdop>\n",
            g_ascii_formatd(sbuf1, sizeof(sbuf1), "%.1f", device->fix->eph / 100.0),
            g_ascii_formatd(sbuf2, sizeof(sbuf2), "%.1f", device->fix->epv));
        g_fprintf(fp, "</trkpt>\n");
      }
    }
  }
  else {
    app_data->has_fix = FALSE;
    app_data->last_device_status = LOCATION_GPS_DEVICE_STATUS_NO_FIX;
  }
	
	//g_print ("Satellites in view: %d\n", device->satellites_in_view);
	//g_print ("Satellites in use: %d\n", device->satellites_in_use);
  CHANGE_LABEL2(app_data->sat_val_label, "%3d/%2d", device->satellites_in_use, device->satellites_in_view);
	//g_print ("GPS status: %d\n", device->status);

	//if (device->cell_info) {
	//	if (device->cell_info->flags & LOCATION_CELL_INFO_GSM_CELL_INFO_SET)
	//		g_print ("Mobile Coutry Code GSM: %d\n", device->cell_info->gsm_cell_info.mcc);

	//	if (device->cell_info->flags & LOCATION_CELL_INFO_WCDMA_CELL_INFO_SET)
	//		g_print ("Mobile Coutry Code WCDMA: %d\n", device->cell_info->wcdma_cell_info.mcc);
	//}

  if(app_data->has_fix) {
    hildon_gtk_window_set_progress_indicator(GTK_WINDOW(app_data->window), 0);
    gtk_widget_set_sensitive(GTK_WIDGET(app_data->start_stop_button), TRUE);
  }
  else {
    hildon_gtk_window_set_progress_indicator(GTK_WINDOW(app_data->window), 1);
    CHANGE_LABEL(app_data->lat_val_label, " ?", NULL);
    CHANGE_LABEL(app_data->lon_val_label, " ?", NULL);
    CHANGE_LABEL(app_data->alt_val_label, " ?", NULL);
    CHANGE_LABEL(app_data->speed_val_label, " ?", NULL);
    CHANGE_LABEL(app_data->track_val_label, " ?", NULL);
    CHANGE_LABEL(app_data->climb_val_label, " ?", NULL);
  }
}

static void
on_gps_error (LocationGPSDevice *device, gpointer data)
{
  //AppData *app_data = data;
	g_error ("GPS error");
}

static void
on_gps_stop (LocationGPSDevice *device, gpointer data)
{
  AppData *app_data = data;
  const gchar *msg = "GPS stopped";
  hildon_banner_show_information(GTK_WIDGET(app_data->window), NULL, msg);
	g_warning (msg);
}

static void
on_gps_start (LocationGPSDevice *device, gpointer data)
{
  AppData *app_data = data;
  const gchar *msg = "GPS started";
  hildon_banner_show_information(GTK_WIDGET(app_data->window), NULL, msg);
	g_warning (msg);
}

int main (int argc, char **argv)
{
  AppData * data = g_new0 (AppData, 1);

	hildon_gtk_init (&argc, &argv);
	LocationGPSDControl *control;
	LocationGPSDevice *device;

  data->working_dir = g_strdup_printf("%s" G_DIR_SEPARATOR_S "MyDocs", g_get_home_dir());
  data->intermediate_gpx_data_filename = g_strdup_printf("%s" G_DIR_SEPARATOR_S ".gps-tracker.gpx", data->working_dir);
	data->program = hildon_program_get_instance ();
	g_set_application_name("GPS Træcker");

	data->window = HILDON_WINDOW(hildon_stackable_window_new());
	hildon_program_add_window (data->program, HILDON_WINDOW (data->window));

  data->wp_marker_str = g_string_sized_new(64);
  data->waypoint_block_str = g_string_sized_new(32<<10);
  data->main_vbox = (gpointer)gtk_vbox_new(FALSE, 0);
  data->loc_hbox = (gpointer)gtk_hbox_new(FALSE, 0);
  data->loc_gps_data_table = (gpointer)gtk_table_new(4, 2, FALSE);
  {
    GtkWidget *alignment_gps = gtk_alignment_new (0.5, 0.5, 1, 1);
    GtkWidget *lat_label = gtk_label_new("Lat:");
    GtkWidget *lon_label = gtk_label_new("Lon:");
    GtkWidget *alt_label = gtk_label_new("Alt:");
    GtkWidget *sat_label = gtk_label_new("Sat:");

    gtk_alignment_set_padding(GTK_ALIGNMENT(alignment_gps), 0,0,0,0);
    gtk_misc_set_alignment (GTK_MISC (lat_label), 1, 0.5);
    gtk_misc_set_alignment (GTK_MISC (lon_label), 1, 0.5);
    gtk_misc_set_alignment (GTK_MISC (alt_label), 1, 0.5);
    gtk_misc_set_alignment (GTK_MISC (sat_label), 1, 0.5);
    gtk_table_attach_defaults(data->loc_gps_data_table, lat_label, 0, 1, 0, 1);
    gtk_table_attach_defaults(data->loc_gps_data_table, lon_label, 0, 1, 1, 2);
    gtk_table_attach_defaults(data->loc_gps_data_table, alt_label, 0, 1, 2, 3);
    gtk_table_attach_defaults(data->loc_gps_data_table, sat_label, 0, 1, 3, 4);
    data->lat_val_label = gtk_label_new(NULL);
    data->lon_val_label = gtk_label_new(NULL);
    data->alt_val_label = gtk_label_new(NULL);
    data->sat_val_label = gtk_label_new(" 0/ 0");
    gtk_misc_set_alignment (GTK_MISC (data->lat_val_label), 0, 0.5);
    gtk_misc_set_alignment (GTK_MISC (data->lon_val_label), 0, 0.5);
    gtk_misc_set_alignment (GTK_MISC (data->alt_val_label), 0, 0.5);
    gtk_misc_set_alignment (GTK_MISC (data->sat_val_label), 0, 0.5);
    gtk_table_attach_defaults(data->loc_gps_data_table, data->lat_val_label, 1, 2, 0, 1);
    gtk_table_attach_defaults(data->loc_gps_data_table, data->lon_val_label, 1, 2, 1, 2);
    gtk_table_attach_defaults(data->loc_gps_data_table, data->alt_val_label, 1, 2, 2, 3);
    gtk_table_attach_defaults(data->loc_gps_data_table, data->sat_val_label, 1, 2, 3, 4);
    data->loc_gps_data = hildon_caption_new(NULL, "", (gpointer)data->loc_gps_data_table, NULL, HILDON_CAPTION_MANDATORY);
    //hildon_caption_set_label_alignment(data->loc_gps_data, 0);
    //hildon_caption_set_separator((gpointer)data->loc_gps_data, "   ");
    hildon_caption_set_label_markup((gpointer)data->loc_gps_data, "<big><b>GPS data</b></big>");
    gtk_container_add (GTK_CONTAINER (alignment_gps), GTK_WIDGET(data->loc_gps_data));
    gtk_box_pack_start(GTK_BOX(data->loc_hbox), alignment_gps, TRUE, FALSE, 0);
  }

  gtk_box_pack_start(GTK_BOX(data->loc_hbox), gtk_vseparator_new(), TRUE, FALSE, 0);

  data->loc_track_data_table = (gpointer)gtk_table_new(4, 2, TRUE);
  {
    GtkWidget *alignment_track = gtk_alignment_new (0.5, 0.5, 1, 1);
    GtkWidget *speed_label = gtk_label_new("Speed:");
    GtkWidget *track_label = gtk_label_new("Orientation:");
    GtkWidget *climb_label = gtk_label_new("Climb:");
    gtk_alignment_set_padding(GTK_ALIGNMENT(alignment_track), 0,0,0,0);
    gtk_misc_set_alignment (GTK_MISC (speed_label), 1, 0.5);
    gtk_misc_set_alignment (GTK_MISC (track_label), 1, 0.5);
    gtk_misc_set_alignment (GTK_MISC (climb_label), 1, 0.5);
    gtk_table_attach_defaults(data->loc_track_data_table, speed_label, 0, 1, 0, 1);
    gtk_table_attach_defaults(data->loc_track_data_table, track_label, 0, 1, 1, 2);
    gtk_table_attach_defaults(data->loc_track_data_table, climb_label, 0, 1, 2, 3);
    data->speed_val_label = gtk_label_new(NULL);
    CHANGE_LABEL(data->speed_val_label, "% 5.1fkm/h", 0.);
    data->track_val_label = gtk_label_new(NULL);
    data->climb_val_label = gtk_label_new(NULL);
    gtk_misc_set_alignment (GTK_MISC (data->speed_val_label), 0, 0.5);
    gtk_misc_set_alignment (GTK_MISC (data->track_val_label), 0, 0.5);
    gtk_misc_set_alignment (GTK_MISC (data->climb_val_label), 0, 0.5);
    gtk_table_attach_defaults(data->loc_track_data_table, data->speed_val_label, 1, 2, 0, 1);
    gtk_table_attach_defaults(data->loc_track_data_table, data->track_val_label, 1, 2, 1, 2);
    gtk_table_attach_defaults(data->loc_track_data_table, data->climb_val_label, 1, 2, 2, 3);
    data->loc_track_data = hildon_caption_new(NULL, "", (gpointer)data->loc_track_data_table, NULL, HILDON_CAPTION_MANDATORY);
    //hildon_caption_set_label_alignment((gpointer)data->loc_track_data, 0);
    hildon_caption_set_label_markup((gpointer)data->loc_track_data, "<big><b>Track</b></big>");
    gtk_container_add (GTK_CONTAINER (alignment_track), GTK_WIDGET(data->loc_track_data));
    gtk_box_pack_end(GTK_BOX(data->loc_hbox), alignment_track, TRUE, FALSE, 0);
  }

  gtk_box_pack_start(GTK_BOX(data->main_vbox), GTK_WIDGET(data->loc_hbox), TRUE, FALSE, 0);

  data->wp_hbox = (gpointer)gtk_hbox_new(TRUE, 0);
  //data->wp_label = gtk_label_new("Marker:");
  data->wp_name_entry = hildon_entry_new (HILDON_SIZE_HALFSCREEN_WIDTH);
  //hildon_entry_set_placeholder (HILDON_ENTRY (data->wp_name_entry), "First name");
  data->wp_set_btn = GTK_BUTTON (hildon_gtk_button_new (HILDON_SIZE_HALFSCREEN_WIDTH));
  gtk_button_set_label (data->wp_set_btn, "Set marker");
  gtk_widget_set_sensitive(GTK_WIDGET(data->wp_set_btn), FALSE);
  //gtk_box_pack_start(GTK_BOX(data->wp_hbox), GTK_WIDGET(data->wp_label), TRUE, FALSE, 0);
  gtk_box_pack_start(GTK_BOX(data->wp_hbox), GTK_WIDGET(data->wp_name_entry), TRUE, FALSE, 0);
  gtk_box_pack_start(GTK_BOX(data->wp_hbox), GTK_WIDGET(data->wp_set_btn), TRUE, FALSE, 0);
  gtk_box_pack_start(GTK_BOX(data->main_vbox), GTK_WIDGET(data->wp_hbox), FALSE, FALSE, 0);

  data->btn_hbox = (gpointer)gtk_hbox_new(TRUE, 0);

  data->start_stop_button = GTK_BUTTON (hildon_gtk_button_new (HILDON_SIZE_THUMB_HEIGHT | HILDON_SIZE_HALFSCREEN_WIDTH));
  gtk_button_set_label (data->start_stop_button, "Start");
  gtk_widget_set_sensitive(GTK_WIDGET(data->start_stop_button), FALSE);
  data->save_button = GTK_BUTTON (hildon_gtk_button_new (HILDON_SIZE_THUMB_HEIGHT | HILDON_SIZE_HALFSCREEN_WIDTH));
  gtk_button_set_label (data->save_button, "Save");
  gtk_widget_set_sensitive(GTK_WIDGET(data->save_button), FALSE);
  gtk_box_pack_start(GTK_BOX(data->btn_hbox), GTK_WIDGET(data->start_stop_button), FALSE, FALSE, 8);
  gtk_box_pack_start(GTK_BOX(data->btn_hbox), GTK_WIDGET(data->save_button), FALSE, FALSE, 8);
  gtk_box_pack_start(GTK_BOX(data->main_vbox), GTK_WIDGET(data->btn_hbox), FALSE, FALSE, 4);

	/* Add vbox to main window */
	gtk_container_add (GTK_CONTAINER (data->window), GTK_WIDGET(data->main_vbox));

  g_signal_connect (G_OBJECT (data->save_button), "clicked", G_CALLBACK (cb_file_save), data);
  g_signal_connect (G_OBJECT (data->start_stop_button), "clicked", G_CALLBACK (cb_start_stop), data);
  g_signal_connect (G_OBJECT (data->wp_set_btn), "clicked", G_CALLBACK (cb_wp_set_btn), data);
	g_signal_connect (G_OBJECT (data->window), "destroy",
			G_CALLBACK (gtk_main_quit), NULL);

	control = location_gpsd_control_get_default ();
  location_gpsd_control_start (control);

	/*
	 * Note that in real life one may want to use some other method and interval
	 * than LOCATION_METHOD_USER_SELECTED and LOCATION_INTERVAL_DEFAULT,
	 * respectively. For more information on possible values for these parameters
	 * please see liblocation online documentation.
	 */
	g_object_set (G_OBJECT (control), 
			"preferred-method", LOCATION_METHOD_USER_SELECTED,
			"preferred-interval", LOCATION_INTERVAL_DEFAULT,
			NULL);

	device  = g_object_new (LOCATION_TYPE_GPS_DEVICE, NULL);

	g_signal_connect (control, "error",		G_CALLBACK (on_gps_error),		data);
	g_signal_connect (control, "gpsd-running",	G_CALLBACK (on_gps_start),		data);
	g_signal_connect (control, "gpsd-stopped",	G_CALLBACK (on_gps_stop),		data);
	g_signal_connect (device,  "changed",		G_CALLBACK (on_gps_device_changed),	data);

	gtk_widget_show_all (GTK_WIDGET (data->window));

  hildon_gtk_window_set_progress_indicator(GTK_WINDOW(data->window), 1);
	gtk_main ();

  location_gpsd_control_stop (control);

  if(data->outf_p)
    fclose(data->outf_p);
  g_unlink(data->intermediate_gpx_data_filename);
  g_free(data->working_dir);
  g_free(data->intermediate_gpx_data_filename);
  g_string_free(data->wp_marker_str, TRUE);
  g_string_free(data->waypoint_block_str, TRUE);
  g_free(data);
	g_object_unref (device);
	g_object_unref (control);

	return 0;
}
